% Newton's method with fixed stepsize 
function [myflag, iter, samples, Hbar, xbar] = GASS_avg(N, K, test_func, M, Hl,initlow,inithigh,  rho, alpha, step, c, choice)
% usage example:
%   [F,I,S,H,X]=GASS_avg(1000,500,@flip_sphere,50,-2e6, -30, 0.05, 1, 0.05, 0.1, 0);
% 
% Input parameters:
% N: number of samples per interation
% K: number of iterations
% test_func: @+function name of the test, a function to be maximized
% M: dimension of the test function
% Hl: lower or higher bound of test_func, meaning that test_func-Hl or Hl-test_func >0
% initlow: lower bound of the range of Optimal Solution in each Dimension
% inithigh: higher bound of the range of Optimal Solution in each Dimension
% rho: Percentage of Elite Samples
% alpha: Initial Step Size
% step: Step Size Parameter
% c: feedback rate
% choice: max(0) or min(1)
%
% Output parameters:
% myflag: has optimal solution or out of maximum iterations or no optimal
% solution or error
% iter: actual number of iterations
% samples: number of samples
% Hbar: optimal value
% xbar: candidate solution


% Initialization
mu_k = rand(1,M)*(inithigh-initlow)+initlow;  % -30 < mu0 < 30
var_k = 1000*ones(1,M);   % the variance vector
theta1_k=mu_k./var_k;
theta2_k=-0.5*ones(1,M)./var_k;
theta_k=[theta1_k, theta2_k];
theta_k_bar=theta_k;

X_k=randn(N,M)*diag(sqrt(var_k))+ones(N,1)*mu_k;
stop=0;
k=1;
myflag = 1;
samples=zeros(K,1);
Hbar = zeros(K,1);
xbar = zeros(K,M);
Ntotal=0;

while (stop==0)&&(k<=K)
    gamma_k=alpha/k^step;  % stepsize
    Ntotal=Ntotal+N;
    samples(k) = Ntotal;
  
    if(choice == 0)
        G = feval(test_func,X_k)-Hl;
    else
        G = -feval(test_func,X_k)+Hl;
    end
    if G<0 % this is a test for whether the user has correctly specified the parameter Hl
        myflag = 0;
        if choice == 0
            w = msgbox('Lower bound is too big, and cannot gaurantee: objective_function - lower_bound >0','Cannot Find Optimal Solution');
        else
            w = msgbox('Higher bound is too small, and cannot gaurantee: higher_bound - objective_function - >0','Cannot Find Optimal Solution');
        end
        waitfor(w);
        break;
    end
    
    % approximate the solution
    [Hbar(k), indx] = max(G);
    xbar(k,:) = X_k(indx,:);

    if isnan(Hbar(k))==1  
        Hbar(k)=Hbar(k-1);
        xbar(k,:) =xbar(k-1,:);
        break;
    end

    if k>2
    if Hbar(k-1)>max(G)     
        Hbar(k)=Hbar(k-1);
        xbar(k,:) =xbar(k-1,:);
    end 
    end

    %--use shape function---take the elite samples---
    [G_sort, I_sort] = sort(G,'descend');   % sorted performances
    gm = G_sort(ceil(N*rho));
    G_u=(G>gm).*G;
    %G_u=G./(1+exp(-1e5*(G-gm)));

    % esitmate gradient and hessian
    w_k=G_u/sum(G_u);
    CX_k=[X_k,X_k.*X_k];
    grad_k=w_k'*CX_k-[mu_k, var_k + mu_k.*mu_k];  % use the exact expectation mu_k instead of the estimate mean(CX_k)
    Hes_k=-cov(CX_k);
    Hes_inv_k=(Hes_k + 1e-30.*eye(2*M))\diag(ones(1,2*M)); 
    %Hes_inv_k=inv(Hes_k + 1e-30.*eye(2*M));  % ensure positive-definiteness
    if norm(grad_k) < 1e-30
        myflag = 1;
        break;
    else
    % update the parameter using an SA iteration
    theta_k=theta_k-gamma_k*(Hes_inv_k*grad_k')'-gamma_k*c*(theta_k_bar-theta_k);
    theta_k_bar=(1-1/(k+1))*theta_k_bar+1/(k+1)*theta_k;
    theta1_k=theta_k(1:M);
    theta2_k=theta_k(M+1:2*M);
    var_k=-0.5./theta2_k;
    mu_k=theta1_k.*var_k;
    % generate new samples
    X_k=randn(N,M)*diag(sqrt(abs(var_k)))+ones(N,1)*mu_k;
    end  %end if

    k=k+1;
    clear G G_u w_k theta1_k theta2_k CX_k grad_k Hes_k Hes_inv_k;
end %end while
iter = k;
if (k<=K) && (myflag == 1)
    myflag = 1;
elseif k>K && myflag == 1
    myflag = 2;
else
    myflag = 0;
end
if k > 2
    k = k-1;
else
    k = 2;
end
for l=k:K
samples(l) = Ntotal;
Hbar(l)=Hbar(k-1);
xbar(l,:) =xbar(k-1,:);
Ntotal=Ntotal+N;
end
if(choice == 0)
    Hbar=Hbar+Hl;
else
    Hbar=-Hbar+Hl;
end